#if __linux || _win32
#include <omp.h>
#endif

#include <maya/MDataBlock.h>
#include <maya/MDataHandle.h>
#include <maya/MPlug.h>
#include <maya/MObject.h>
#include <maya/MGlobal.h>
#include <maya/MTypeId.h>
#include <maya/MFnNumericAttribute.h>
#include <maya/MFnGenericAttribute.h>
#include <maya/MFnTypedAttribute.h>
#include <maya/MFnEnumAttribute.h>
#include <maya/MFnMesh.h>
#include <maya/MFnMeshData.h>
#include <maya/MIntArray.h>
#include <maya/MFloatPoint.h>
#include <maya/MFloatVector.h>
#include <maya/MVectorArray.h>
#include <maya/MFloatVectorArray.h>
#include <maya/MFnVectorArrayData.h>
#include <maya/MFloatPointArray.h>
#include <maya/MItMeshVertex.h>
#include <maya/MMeshIntersector.h>
#include <maya/MFloatMatrix.h>
#include <maya/MPlugArray.h>
#include <maya/MSelectionList.h>
#include <maya/MItMeshEdge.h>

#include "tetrahedralize.h"
#include "../pluginMain.h"

#include "../wm5/Wm5Delaunay3.h"
#include "../wm5/Wm5Vector3.h"
#include "../wm5/Wm5Delaunay2.h"
#include "../wm5/Wm5Vector2.h"
#include "../wm5/Wm5Math.h"



// tetgen
#define TETLIBRARY
#include "tetgen1.5.0/tetgen.h"

#define DEFAULT_FULL_DEPTH 5

#define XSTR(x) STR(x)
#define STR(x) #x
#if DEFAULT_FULL_DEPTH
#pragma message ( "[WARNING] Setting default full depth to " XSTR(DEFAULT_FULL_DEPTH) )
#endif // DEFAULT_FULL_DEPTH
#include <stdarg.h>



// cgal
#include <CGAL/Polygon_mesh_processing/self_intersections.h>
#include <CGAL/Exact_predicates_inexact_constructions_kernel.h>
#include <CGAL/Surface_mesh.h>
#include <fstream>
#include <vector>

typedef CGAL::Exact_predicates_inexact_constructions_kernel K;
typedef CGAL::Surface_mesh<K::Point_3> Mesh;
typedef Mesh::Vertex_index vertex_descriptor;
typedef boost::graph_traits<Mesh>::face_descriptor face_descriptor;
namespace PMP = CGAL::Polygon_mesh_processing;



MTypeId tetrahedralize::id(ID_TETRAHEDRALIZE);
MObject tetrahedralize::aInMesh;
MObject tetrahedralize::aInPosPP;
MObject tetrahedralize::aInNmlPP;
MObject tetrahedralize::aOutMesh;
MObject tetrahedralize::aActive;
MObject tetrahedralize::aMode;
MObject tetrahedralize::aUseMsh;
MObject tetrahedralize::aUseCld;
MObject tetrahedralize::aUseVol;
MObject tetrahedralize::aMaxVol;
MObject tetrahedralize::aUseRad;
MObject tetrahedralize::aMaxRad;
MObject tetrahedralize::aIgnBndFcs;
MObject tetrahedralize::aInhNml;

tetrahedralize::tetrahedralize(){}
tetrahedralize::~tetrahedralize(){}

void* tetrahedralize::creator()
{
	return new tetrahedralize();
}

MStatus tetrahedralize::initialize()
{
	MFnNumericAttribute nAttr;
	MFnGenericAttribute gAttr;
	MFnTypedAttribute tAttr;
	MFnEnumAttribute eAttr;

	aInMesh = gAttr.create("inMesh", "im");
	gAttr.addDataAccept(MFnData::kMesh);
	gAttr.setStorable(false);
	gAttr.setDisconnectBehavior(MFnAttribute::kReset);

	aInPosPP = gAttr.create("inPositionPP", "ippp");
	gAttr.addDataAccept(MFnData::kVectorArray);
	gAttr.setStorable(false);
	gAttr.setDisconnectBehavior(MFnAttribute::kReset);

	aInNmlPP = gAttr.create("inNormalPP", "inpp");
	gAttr.addDataAccept(MFnData::kVectorArray);
	gAttr.setStorable(false);
	gAttr.setDisconnectBehavior(MFnAttribute::kReset);

	aOutMesh = gAttr.create("outMesh", "om");
	gAttr.addDataAccept(MFnData::kMesh);
	gAttr.setWritable(false);
	gAttr.setStorable(false);

	aActive = nAttr.create("active", "act", MFnNumericData::kBoolean, true);
	nAttr.setKeyable(true);

	aMode = eAttr.create("mode", "mode", MFnNumericData::kShort);
	eAttr.addField("tetrahedralize", 0);
	eAttr.addField("cellularize", 1);
	eAttr.addField("triangulate", 2);
	eAttr.addField("convex hull", 3);
	eAttr.setDefault(0);
	eAttr.setKeyable(true);

	aUseMsh = nAttr.create("useBoundaryMesh", "ubm", MFnNumericData::kBoolean, true);
	nAttr.setKeyable(true);

	aUseCld = nAttr.create("usePointCloud", "upc", MFnNumericData::kBoolean, true);
	nAttr.setKeyable(true);

	aUseVol = nAttr.create("useSizeLimit", "usl", MFnNumericData::kBoolean, false);
	nAttr.setKeyable(true);

	aMaxVol = nAttr.create("size", "sz", MFnNumericData::kFloat, 1.0);
	nAttr.setKeyable(true);
	nAttr.setMin(0.0);
	nAttr.setSoftMax(2.0);

	aUseRad = nAttr.create("useRefine", "urfn", MFnNumericData::kBoolean, false);
	nAttr.setKeyable(true);

	aMaxRad = nAttr.create("refine", "rfn", MFnNumericData::kFloat, 1.0);
	nAttr.setKeyable(true);
	nAttr.setMin(0.0);
	nAttr.setMax(1.0);

	aIgnBndFcs = nAttr.create("ignoreBoundaryFaces", "ibf", MFnNumericData::kBoolean, false);
	nAttr.setKeyable(true);

	aInhNml = nAttr.create("inheritNormals", "inmls", MFnNumericData::kBoolean, true);
	nAttr.setKeyable(true);

	addAttribute(aInMesh);
	addAttribute(aInPosPP);
	addAttribute(aInNmlPP);
	addAttribute(aOutMesh);
	addAttribute(aActive);
	addAttribute(aMode);
	addAttribute(aUseMsh);
	addAttribute(aUseCld);
	addAttribute(aUseVol);
	addAttribute(aMaxVol);
	addAttribute(aUseRad);
	addAttribute(aMaxRad);
	addAttribute(aIgnBndFcs);
	addAttribute(aInhNml);

	attributeAffects(aInMesh, aOutMesh);
	attributeAffects(aInPosPP, aOutMesh);
	attributeAffects(aInNmlPP, aOutMesh);
	attributeAffects(aActive, aOutMesh);
	attributeAffects(aMode, aOutMesh);
	attributeAffects(aUseMsh, aOutMesh);
	attributeAffects(aUseCld, aOutMesh);
	attributeAffects(aUseVol, aOutMesh);
	attributeAffects(aMaxVol, aOutMesh);
	attributeAffects(aUseRad, aOutMesh);
	attributeAffects(aMaxRad, aOutMesh);
	attributeAffects(aIgnBndFcs, aOutMesh);
	attributeAffects(aInhNml, aOutMesh);

	return MS::kSuccess;
}

MStatus tetrahedralize::compute(const MPlug& plug, MDataBlock& data)
{
	if(data.outputValue(state).asShort() == 2) return MS::kUnknownParameter;

	if(plug != aOutMesh) return MS::kUnknownParameter;

	bool bUseMsh = data.inputValue(aUseMsh).asBool();
	bool bUseCld = data.inputValue(aUseCld).asBool();

	//
	// input data
	//

	// mesh
	MFnMeshData md;
	MObject o = md.create();
	MFnMesh m;
	m.copy(data.inputValue(aInMesh).asMesh(), o);
	m.setObject(o);

	if(data.inputValue(aActive).asBool() == false)
	{
		MFnMesh m2;
		MFnMeshData md2;
		o = md2.create();
		m2.create(0, 0, MFloatPointArray(), MIntArray(), MIntArray(), o);
		data.outputValue(aOutMesh).set(o);
		return MS::kSuccess;
	}

	int nPntCnt = m.numVertices();
	if(nPntCnt == 0) bUseMsh = false;

	// vertices
	MFloatPointArray pnts;
	if(bUseMsh) m.getPoints(pnts, MSpace::kWorld);
	else nPntCnt = 0;

	short mode = data.inputValue(aMode).asShort();
	if(mode < 0) mode = 0;
	else if(mode > 3) mode = 3;

	// this node's name
	MStringArray sa;
	MPlug(thisMObject(), aInMesh).name().split('.', sa);

	MPointOnMesh pomi;
	MMeshIntersector meshIntersector;
	if(bUseMsh) meshIntersector.create(o);

	// point cloud
	int nPcdCnt = 0;
	if(bUseCld)
	{
		MVectorArray a;
		MObject o2 = data.inputValue(aInPosPP).data();
		a = MFnVectorArrayData(o2).array();
		nPcdCnt = a.length();

		if(!bUseMsh) pnts.setLength(nPntCnt+nPcdCnt);

		MFloatPoint pnt;
		for(int i = 0; i < nPcdCnt; ++i)
		{
			if(!bUseMsh) pnts[nPntCnt+i] = a[i];
			else
			{
				meshIntersector.getClosestPoint(a[i], pomi, 999999999.9);
				if(pomi.getPoint().distanceTo(a[i]) <= 0.0001) continue;
				pnts.append(a[i]);
			}
		}

		if(!bUseMsh) nPntCnt += nPcdCnt;
		else
		{
			int cnt = pnts.length();
			nPcdCnt = cnt - nPntCnt;
			nPntCnt = cnt;
		}
	}

	if(nPntCnt == 0)
	{
		MFnMesh m;
		MFnMeshData md;
		o = md.create();
		m.create(0, 0, MFloatPointArray(), MIntArray(), MIntArray(), o);
		data.outputValue(aOutMesh).set(o);
		return MS::kSuccess;
	}

	if(mode == 3)
	{
		mode = 2;
		bUseMsh = false;
		bUseCld = true;
		nPcdCnt = 0;
	}

	// check dimension of mesh + point cloud
	Wm5::Vector3f *v = (Wm5::Vector3f*)malloc(sizeof(Wm5::Vector3f)*nPntCnt);
	for(int i = 0; i < nPntCnt; ++i) v[i] = Wm5::Vector3f(pnts[i].x, pnts[i].y, pnts[i].z);
	Wm5::Delaunay3f del3(nPntCnt, v, 0.0001f, false, Wm5::Query::QT_INT64);
	int dim = del3.GetDimension();
	free(v);
	if(dim == 3)
	{
		if(bUseMsh)
		{
			// check for holes in the mesh
			MItMeshVertex it(o);
			for(; !it.isDone(); it.next())
				if(it.onBoundary())
				{
					MGlobal::displayWarning(sa[0]+": Cannot proceed - input mesh has hole(s).");
					data.outputValue(aOutMesh).set(o);
					return MS::kSuccess;
				}

			// check for flat mesh
			Wm5::Vector3f *v = (Wm5::Vector3f*)malloc(sizeof(Wm5::Vector3f)*nPntCnt-nPcdCnt);
			for(int i = 0; i < nPntCnt-nPcdCnt; ++i) v[i] = Wm5::Vector3f(pnts[i].x, pnts[i].y, pnts[i].z);
			Wm5::Delaunay3f del3(nPntCnt-nPcdCnt, v, 0.0001f, false, Wm5::Query::QT_INT64);
			dim = del3.GetDimension();
			free(v);
			if(dim != 3)
			{
				MGlobal::displayWarning(sa[0]+": Cannot proceed - input mesh is flat.");
				data.outputValue(aOutMesh).set(o);
				return MS::kSuccess;
			}
		}

		// check for flat point cloud
		if(bUseCld && nPcdCnt > 0 && !bUseMsh)
		{
			Wm5::Vector3f *v = (Wm5::Vector3f*)malloc(sizeof(Wm5::Vector3f)*nPcdCnt);
			for(int i = nPntCnt-nPcdCnt; i < nPntCnt; ++i) v[i] = Wm5::Vector3f(pnts[i].x, pnts[i].y, pnts[i].z);
			Wm5::Delaunay3f del3(nPcdCnt, v, 0.0001f, false, Wm5::Query::QT_INT64);
			dim = del3.GetDimension();
			free(v);
			if(dim != 3)
			{
				MGlobal::displayWarning(sa[0]+": input point cloud is flat.");
				data.outputValue(aOutMesh).set(o);
				return MS::kSuccess;
			}
		}
	}
	else if(dim == 2)
	{
		Wm5::Vector3f dir = del3.GetPlaneDirection(0);
		Wm5::Vector3f dir2 = del3.GetPlaneDirection(1);
		Wm5::Vector3f orig = del3.GetPlaneOrigin();
		MFloatVector vx, vz, vy;
		vx.x = dir[0]; vx.y = dir[1]; vx.z = dir[2]; vx.normalize();
		vz.x = dir2[0]; vz.y = dir2[1]; vz.z = dir2[2]; vz.normalize();
		vy = vx^vz; vy.normalize();
		vz = vx^vy; vz.normalize();
		MFloatMatrix m;
		m.matrix[0][0] = vx.x;
		m.matrix[0][1] = vx.y;
		m.matrix[0][2] = vx.z;
		m.matrix[1][0] = vy.x;
		m.matrix[1][1] = vy.y;
		m.matrix[1][2] = vy.z;
		m.matrix[2][0] = vz.x;
		m.matrix[2][1] = vz.y;
		m.matrix[2][2] = vz.z;
		m.matrix[3][0] = orig[0];
		m.matrix[3][1] = orig[1];
		m.matrix[3][2] = orig[2];
		Wm5::Vector2f *v = (Wm5::Vector2f*)malloc(sizeof(Wm5::Vector2f)*nPntCnt);
		for(int i = 0; i < nPntCnt; ++i)
		{
			vx = pnts[i] * m.inverse();
			v[i] = Wm5::Vector2f(vx.x, vx.z);
		}

		Wm5::Delaunay2f del2(nPntCnt, v, 0.0001f, false, Wm5::Query::QT_INT64);
		int c = del2.GetNumSimplices();
		if(c > 0)
		{
			const int *ids = del2.GetIndices();
			MIntArray cnts(c, 3);
			MIntArray conn(c*3);
			for(int i = 0; i < c; i++)
			{
				conn[i*3] = ids[i*3];
				conn[i*3+1] = ids[i*3+1];
				conn[i*3+2] = ids[i*3+2];
			}

			free(v);
			MFnMesh m;
			MFnMeshData md;
			o = md.create();
			m.create(nPntCnt, cnts.length(), pnts, cnts, conn, o);
			data.outputValue(aOutMesh).set(o);
			return MS::kSuccess;
		}
		else
		{
			free(v);
			data.outputValue(aOutMesh).set(o);
			return MS::kSuccess;
		}
	}
	else
	{
		MGlobal::displayWarning(sa[0]+": Cannot proceed - input mesh and point cloud lie on a line.");
		data.outputValue(aOutMesh).set(o);
		return MS::kSuccess;
	}

	MIntArray cnts, conn;
	int j, j1, j2, j3;
	int cnt = 0;

	if(bUseMsh)
	{
		m.getTriangles(cnts, conn);

		// check for self-intersecting mesh
		//
		Mesh mesh;
		std::vector<vertex_descriptor> vd(nPntCnt);
		for(unsigned int i = 0; i < nPntCnt; ++i)
			vd[i] = mesh.add_vertex(K::Point_3(pnts[i].x, pnts[i].y, pnts[i].z));
		for(unsigned int i = 0; i < cnts.length(); ++i)
			mesh.add_face(vd[conn[i*3]], vd[conn[i*3+1]], vd[conn[i*3+2]]);
		std::vector<std::pair<face_descriptor, face_descriptor> > facets;
		PMP::self_intersections(faces(mesh), mesh, std::back_inserter(facets));
		if(!facets.empty())
		{
			MGlobal::displayWarning(sa[0]+": Cannot proceed - input mesh is self-intersecting. Check console for list of intersecting faces.");
#ifdef __APPLE__
			cout<<"Intersecting faces:"<<endl;
#else
			cout<<sa[0]+": Intersecting faces:"<<endl;
#endif
			for(unsigned int i = 0; i < facets.size(); i++) cout<<"  "<<facets[i].first<<" - "<<facets[i].second<<endl;
			data.outputValue(aOutMesh).set(o);
			return MS::kSuccess;
		}
		//
		// END check for self-intersections
	}

	//
	// output data
	//

	tetgenio in, out;
	in.firstnumber = 1; // All indices start from 1.

	// input points
	in.numberofpoints = nPntCnt;
	in.pointlist = new REAL[nPntCnt*3];
	for(unsigned int i = 0; i < nPntCnt; ++i)
	{
		j = i*3;
		in.pointlist[j] = pnts[i].x;
		in.pointlist[j+1] = pnts[i].y;
		in.pointlist[j+2] = pnts[i].z;
	}

	// mesh boundary
	if(bUseMsh)
	{
		tetgenio::facet *f;
		tetgenio::polygon *p;
		in.numberoffacets = conn.length()/3;
		in.facetlist = new tetgenio::facet[in.numberoffacets];
		for(int i = 0; i < in.numberoffacets; ++i)
		{
			f = &in.facetlist[i];
			f->numberofpolygons = 1;
			f->polygonlist = new tetgenio::polygon[f->numberofpolygons];
			f->numberofholes = 0;
			f->holelist = NULL;
			p = &f->polygonlist[0];
			p->numberofvertices = 3;
			p->vertexlist = new int[p->numberofvertices];
			j = i*3;
			p->vertexlist[0] = conn[j]+1;
			p->vertexlist[1] = conn[j+1]+1;
			p->vertexlist[2] = conn[j+2]+1;
		}
	}

	bool bUseVol = data.inputValue(aUseVol).asBool();
	if(bUseMsh)
	{
		MString s1;
		if(data.inputValue(aUseRad).asBool())
		{
			float fMaxRad = data.inputValue(aMaxRad).asFloat()*0.9 + 1.1;
			if(fMaxRad < 1.1) fMaxRad = 1.1;
			else if(fMaxRad > 2.0) fMaxRad = 2.0;
			s1 = "q";
			if(fMaxRad < 2.0) s1 += fMaxRad;
		}

		MString s2;
		if(bUseVol)
		{
			float fMaxVol = (data.inputValue(aMaxVol).asFloat()+0.1) * 0.01;
			if(fMaxVol < 0.001) fMaxVol = 0.001;
			s2 = "a";
			s2 += fMaxVol;
		}

		_tetrahedralize((char *)(("Qnp"+s1+s2).asChar()), &in, &out);
	}
	else
		_tetrahedralize("Qn", &in, &out);

	nPntCnt = out.numberofpoints;
	if(nPntCnt == 0)
	{
		data.outputValue(aOutMesh).set(o);
		return MS::kSuccess;
	}

	pnts.setLength(nPntCnt);
	for(int i = 0; i < nPntCnt; ++i)
	{
		j = i*3;
		pnts[i].x = out.pointlist[j];
		pnts[i].y = out.pointlist[j+1];
		pnts[i].z = out.pointlist[j+2];
		pnts[i].w = 1.0;
	}

	MIntArray ids;
	MVectorArray nmls;
	bool bInhNml = data.inputValue(aInhNml).asBool();
	if(bInhNml && bUseMsh && mode != 1)
	{
		ids.setLength(nPntCnt);
		nmls.setLength(nPntCnt);
		for(int i = 0; i < nPntCnt; ++i)
		{
			meshIntersector.getClosestPoint(pnts[i], pomi, 999999999.9);
			nmls[i] = pomi.getNormal();
			ids[i] = i;
		}
	}

	bool bIgnBndFcs = data.inputValue(aIgnBndFcs).asBool();
	int jj, jj1, jj2, jj3;
	if(mode == 0)
	{
		cnt = 0;
		cnts.clear();
		conn.clear();
		for(int i = 0; i < out.numberoftetrahedra; ++i)
		{
			j = i*4;
			j1 = j+1;
			j2 = j+2;
			j3 = j+3;
			jj = out.tetrahedronlist[j]-1;
			jj1 = out.tetrahedronlist[j1]-1;
			jj2 = out.tetrahedronlist[j2]-1;
			jj3 = out.tetrahedronlist[j3]-1;

			if(out.neighborlist[j3] > i || (out.neighborlist[j3] == -1 && bIgnBndFcs == false))
			{
				conn.setLength(cnt+3);
				conn[cnt] = jj;
				conn[cnt+1] = jj1;
				conn[cnt+2] = jj2;
				cnt += 3;
			}
			if(out.neighborlist[j2] > i || (out.neighborlist[j2] == -1 && bIgnBndFcs == false))
			{
				conn.setLength(cnt+3);
				conn[cnt] = jj;
				conn[cnt+1] = jj1;
				conn[cnt+2] = jj3;
				cnt += 3;
			}
			if(out.neighborlist[j] > i || (out.neighborlist[j] == -1 && bIgnBndFcs == false))
			{
				conn.setLength(cnt+3);
				conn[cnt] = jj1;
				conn[cnt+1] = jj2;
				conn[cnt+2] = jj3;
				cnt += 3;
			}
			if(out.neighborlist[j1] > i || (out.neighborlist[j1] == -1 && bIgnBndFcs == false))
			{
				conn.setLength(cnt+3);
				conn[cnt] = jj;
				conn[cnt+1] = jj2;
				conn[cnt+2] = jj3;
				cnt += 3;
			}
		}
		cnt /= 3;
		cnts = MIntArray(cnt, 3);
	}
	else if(mode == 1)
	{
		int c01, c02, c03, c12, c13, c23, b0123, b012, b013, b023, b123;
		nPntCnt = out.numberoftetrahedra*11;
		cnt = 0;
		cnts.clear();
		conn.clear();
		MFloatPointArray pnts2(nPntCnt);
		for(int i = 0; i < out.numberoftetrahedra; ++i)
		{
			j = i*4;
			j1 = j+1;
			j2 = j+2;
			j3 = j+3;
			jj = out.tetrahedronlist[j]-1;
			jj1 = out.tetrahedronlist[j1]-1;
			jj2 = out.tetrahedronlist[j2]-1;
			jj3 = out.tetrahedronlist[j3]-1;

			c01 = i*11;
			c02 = i*11+1;
			c03 = i*11+2;
			c12 = i*11+3;
			c13 = i*11+4;
			c23 = i*11+5;
			b0123 = i*11+6;
			b012 = i*11+7;
			b013 = i*11+8;
			b023 = i*11+9;
			b123 = i*11+10;

			pnts2[c01] = (pnts[jj]+pnts[jj1])*0.5;
			pnts2[c02] = (pnts[jj]+pnts[jj2])*0.5;
			pnts2[c03] = (pnts[jj]+pnts[jj3])*0.5;
			pnts2[c12] = (pnts[jj1]+pnts[jj2])*0.5;
			pnts2[c13] = (pnts[jj1]+pnts[jj3])*0.5;
			pnts2[c23] = (pnts[jj2]+pnts[jj3])*0.5;
			pnts2[b0123] = (pnts[jj]+pnts[jj1]+pnts[jj2]+pnts[jj3])*0.25;
			pnts2[b012] = (pnts[jj]+pnts[jj1]+pnts[jj2])*0.333333333;
			pnts2[b013] = (pnts[jj]+pnts[jj1]+pnts[jj3])*0.333333333;
			pnts2[b023] = (pnts[jj]+pnts[jj2]+pnts[jj3])*0.333333333;
			pnts2[b123] = (pnts[jj1]+pnts[jj2]+pnts[jj3])*0.333333333;

			conn.setLength(cnt+24);
			conn[cnt] = b0123;
			conn[cnt+1] = b012;
			conn[cnt+2] = c01;
			conn[cnt+3] = b013;

			conn[cnt+4] = b0123;
			conn[cnt+5] = b012;
			conn[cnt+6] = c12;
			conn[cnt+7] = b123;

			conn[cnt+8] = b0123;
			conn[cnt+9] = b012;
			conn[cnt+10] = c02;
			conn[cnt+11] = b023;

			conn[cnt+12] = b0123;
			conn[cnt+13] = b013;
			conn[cnt+14] = c03;
			conn[cnt+15] = b023;

			conn[cnt+16] = b0123;
			conn[cnt+17] = b013;
			conn[cnt+18] = c13;
			conn[cnt+19] = b123;

			conn[cnt+20] = b0123;
			conn[cnt+21] = b023;
			conn[cnt+22] = c23;
			conn[cnt+23] = b123;

			cnt += 24;

			if(out.neighborlist[j3] == -1 && bIgnBndFcs == false)
			{
				pnts2.setLength(nPntCnt+3);
				conn.setLength(cnt+12);

				pnts2[nPntCnt] = pnts[jj];
				conn[cnt] = nPntCnt;
				conn[cnt+1] = c01;
				conn[cnt+2] = b012;
				conn[cnt+3] = c02;

				pnts2[nPntCnt+1] = pnts[jj1];
				conn[cnt+4] = nPntCnt+1;
				conn[cnt+5] = c01;
				conn[cnt+6] = b012;
				conn[cnt+7] = c12;

				pnts2[nPntCnt+2] = pnts[jj2];
				conn[cnt+8] = nPntCnt+2;
				conn[cnt+9] = c12;
				conn[cnt+10] = b012;
				conn[cnt+11] = c02;

				nPntCnt += 3;
				cnt += 12;
			}
			if(out.neighborlist[j2] == -1 && bIgnBndFcs == false)
			{
				pnts2.setLength(nPntCnt+3);
				conn.setLength(cnt+12);

				pnts2[nPntCnt] = pnts[jj];
				conn[cnt] = nPntCnt;
				conn[cnt+1] = c01;
				conn[cnt+2] = b013;
				conn[cnt+3] = c03;

				pnts2[nPntCnt+1] = pnts[jj1];
				conn[cnt+4] = nPntCnt+1;
				conn[cnt+5] = c01;
				conn[cnt+6] = b013;
				conn[cnt+7] = c13;

				pnts2[nPntCnt+2] = pnts[jj3];
				conn[cnt+8] = nPntCnt+2;
				conn[cnt+9] = c03;
				conn[cnt+10] = b013;
				conn[cnt+11] = c13;

				nPntCnt += 3;
				cnt += 12;
			}
			if(out.neighborlist[j] == -1 && bIgnBndFcs == false)
			{
				pnts2.setLength(nPntCnt+3);
				conn.setLength(cnt+12);

				pnts2[nPntCnt] = pnts[jj1];
				conn[cnt] = nPntCnt;
				conn[cnt+1] = c12;
				conn[cnt+2] = b123;
				conn[cnt+3] = c13;

				pnts2[nPntCnt+1] = pnts[jj2];
				conn[cnt+4] = nPntCnt+1;
				conn[cnt+5] = c12;
				conn[cnt+6] = b123;
				conn[cnt+7] = c23;

				pnts2[nPntCnt+2] = pnts[jj3];
				conn[cnt+8] = nPntCnt+2;
				conn[cnt+9] = c13;
				conn[cnt+10] = b123;
				conn[cnt+11] = c23;

				nPntCnt += 3;
				cnt += 12;
			}
			if(out.neighborlist[j1] == -1 && bIgnBndFcs == false)
			{
				pnts2.setLength(nPntCnt+3);
				conn.setLength(cnt+12);

				pnts2[nPntCnt] = pnts[jj];
				conn[cnt] = nPntCnt;
				conn[cnt+1] = c02;
				conn[cnt+2] = b023;
				conn[cnt+3] = c03;

				pnts2[nPntCnt+1] = pnts[jj2];
				conn[cnt+4] = nPntCnt+1;
				conn[cnt+5] = c02;
				conn[cnt+6] = b023;
				conn[cnt+7] = c23;

				pnts2[nPntCnt+2] = pnts[jj3];
				conn[cnt+8] = nPntCnt+2;
				conn[cnt+9] = c03;
				conn[cnt+10] = b023;
				conn[cnt+11] = c23;

				nPntCnt += 3;
				cnt += 12;
			}
		}
		cnt /= 4;
		cnts = MIntArray(cnt, 4);
		pnts = pnts2;
	}
	else if(mode == 2)
	{
		cnt = out.numberoftrifaces;
		cnts = MIntArray(cnt, 3);
		conn.setLength(cnt * 3);
		for(int i = 0; i < cnt; ++i)
		{
			j = i*3;
			j1 = j+1;
			j2 = j+2;

			conn[j] = out.trifacelist[j]-1;
			conn[j1] = out.trifacelist[j1]-1;
			conn[j2] = out.trifacelist[j2]-1;
		}
	}

	//
	// create output mesh
	//
	MFnMesh m2;
	MFnMeshData md2;
	o = md2.create();
	m2.create(nPntCnt, cnt, pnts, cnts, conn, o);
	if(bInhNml && bUseMsh) m2.setVertexNormals(nmls, ids, MSpace::kWorld);
	data.outputValue(aOutMesh).set(o);

	return MS::kSuccess;
}
